import time
from copy import deepcopy
from types import MappingProxyType

from mona_sdk import MonaSingleMessage

from .endpoints.wrapping_getter import get_endpoint_wrapping
from .mona_client import get_mona_clients
from .util.func_util import add_conditional_sampling
from .util.async_util import (
    run_in_an_event_loop,
    call_non_blocking_sync_or_async,
)
from .util.openai_util import get_model_param
from .util.tokens_util import get_usage
from .util.stream_util import ResponseGatheringIterator
from .util.validation_util import validate_and_get_sampling_ratio

EMPTY_DICT = MappingProxyType({})

MONA_ARGS_PREFIX = "MONA_"
CONTEXT_ID_ARG_NAME = MONA_ARGS_PREFIX + "context_id"
EXPORT_TIMESTAMP_ARG_NAME = MONA_ARGS_PREFIX + "export_timestamp"
ADDITIONAL_DATA_ARG_NAME = MONA_ARGS_PREFIX + "additional_data"


def _get_mona_single_message(
    api_name,
    request_input,
    start_time,
    is_exception,
    is_async,
    stream_start_time,
    response,
    analysis_getter,
    context_class,
    message_cleaner,
    additional_data,
    context_id,
    export_timestamp,
):
    """
    Returns a MonaSingleMessage object to be used for data
    exporting to Mona's servers by a Mona client.
    """

    message = {
        "input": request_input,
        "latency": time.time() - start_time,
        "stream_start_latency": stream_start_time - start_time
        if stream_start_time is not None
        else None,
        "is_exception": is_exception,
        "api_name": api_name,
        "is_async": is_async,
    }

    if additional_data:
        message["additional_data"] = additional_data

    if response:
        message["response"] = response
        message["analysis"] = analysis_getter(request_input, response)

    message = message_cleaner(message)

    return MonaSingleMessage(
        message=message,
        contextClass=context_class,
        contextId=context_id,
        exportTimestamp=export_timestamp,
    )


# TODO(itai): Consider creating some sturct (as NamedTuple or dataclass) for
#   the specs param.


def monitor(
    openai_class,
    mona_creds,
    context_class,
    specs=EMPTY_DICT,
    mona_clients_getter=get_mona_clients,
):
    """
    Returns a Wrapped version of a given OpenAI class with mona
    monitoring logic.
    This is the main exposed function of the mona_openai package and
    probably the only thing you need to use from this package.

    You can use the returned class' "create" and "acreate" functions
    exactly as you would the original class, and monitoring will be
    taken care of for you.

    This client will automatically monitor for you things like latency,
    prompt and response lengths, number of tokens, etc., along with any
    endpoint parameter usage (e.g., it tracks the "temperature" and
    "max_tokens" params you use).

    You can also add other named args when calling "create" or
    "acreate" by using a new named argument called
    "MONA_additional_data" and set it to any JSON serializable
    dictionary.
    This allows you to add metadata about the call such as a prompt
    template ID, information about the context in which the API call is
    made, etc...

    Furthermore, you can add to create/acreate functions mona specific
    arguments:
        MONA_context_id: The unique id of the context in which the call
            is made. By using this ID you can export more data to Mona
            to the same context from other places. If not used, the
            "id" field of the OpenAI Endpoint's response will be used.
        MONA_export_timestamp: Can be used to simulate as if the
            current call was made in a different time, as far as Mona
            is concerned.

    The returned monitored class also exposes a new class method called
    "get_mona_clients", which allows you to retrieve the clients and
    use them to export more data or communicate directly with Mona's
    API

    Read more about Mona and how to use it in Mona's docs on
    https://docs.monalabs.io.

    Args:
        openai_class: An OpenAI API class to wrap with monitoring
            capabilties.
        mona_creds: Either a dict or pair of Mona API key and secret to
            set up Mona's clients from its SDK
        context_class: The Mona context class name to use for
            monitoring. Use a name of your choice.
        specs: A dictionary of specifications such as monitoring
            sampling ratio.
        mona_clients_getter: Used only for testing purposes
    """
    client, async_client = mona_clients_getter(mona_creds)

    sampling_ratio = validate_and_get_sampling_ratio(specs)

    base_class = get_endpoint_wrapping(
        openai_class.__name__, specs
    ).wrap_class(openai_class)

    # TODO(itai): Add call to Mona servers to init the context class if it
    #   isn't inited yet once we have the relevant endpoint for this.

    class MonitoredOpenAI(base_class):
        """
        A mona-monitored version of an openai API class.
        """

        @classmethod
        def _get_mona_single_message(
            cls,
            kwargs_param,
            start_time,
            is_exception,
            is_async,
            stream_start_time,
            response,
        ):
            """
            Returns a MonaSingleMessage object to be used for data
            exporting to Mona's servers by a Mona client.
            """
            # Recreate the input dict to avoid manipulating the caller's data,
            # and remove Mona-related data.
            request_input = deepcopy(
                {
                    x: kwargs_param[x]
                    for x in kwargs_param
                    if not x.startswith(MONA_ARGS_PREFIX)
                }
            )

            return _get_mona_single_message(
                api_name=openai_class.__name__,
                request_input=request_input,
                start_time=start_time,
                is_exception=is_exception,
                is_async=is_async,
                stream_start_time=stream_start_time,
                response=response,
                analysis_getter=super()._get_full_analysis,
                context_class=context_class,
                message_cleaner=super()._get_clean_message,
                additional_data=kwargs_param.get(ADDITIONAL_DATA_ARG_NAME),
                context_id=kwargs_param.get(
                    CONTEXT_ID_ARG_NAME, response["id"] if response else None
                ),
                export_timestamp=kwargs_param.get(
                    EXPORT_TIMESTAMP_ARG_NAME, start_time
                ),
            )

        @classmethod
        async def _inner_create(
            cls, export_function, super_function, args, kwargs
        ):
            """
            The main logic for wrapping create functions with mona data
            exporting.
            This internal function porovides a template for both sync
            and async activations (helps with wrapping both "create"
            and "acreate").
            """

            is_stream = kwargs.get("stream", False)
            is_async = super_function.__name__ == "acreate"

            response = None

            # will be used only when stream is enabled
            stream_start_time = None

            async def _inner_mona_export(is_exception):
                return await call_non_blocking_sync_or_async(
                    export_function,
                    (
                        cls._get_mona_single_message(
                            kwargs,
                            start_time,
                            is_exception,
                            is_async,
                            stream_start_time,
                            response,
                        ),
                    ),
                )

            mona_export = add_conditional_sampling(
                _inner_mona_export, sampling_ratio
            )

            start_time = time.time()

            async def inner_super_function():
                # Call the actual openai create function without the Mona
                # specific arguments.
                return await call_non_blocking_sync_or_async(
                    super_function,
                    args,
                    {
                        x: kwargs[x]
                        for x in kwargs
                        if not x.startswith(MONA_ARGS_PREFIX)
                    },
                )

            async def inner_handle_exception():
                if not specs.get("avoid_monitoring_exceptions", False):
                    await mona_export(True)

            if not is_stream:
                try:
                    response = await inner_super_function()
                except Exception:
                    await inner_handle_exception()
                    raise

                await mona_export(False)

                return response

            # From here it's stream handling.

            async def _stream_done_callback(
                final_response, actual_stream_start_time
            ):
                nonlocal response
                nonlocal stream_start_time
                # There is no usage data in returned stream responses, so
                # we add it here.
                response = final_response | {
                    "usage": get_usage(
                        model=get_model_param(kwargs),
                        prompt_texts=base_class._get_all_prompt_texts(kwargs),
                        response_texts=base_class._get_all_response_texts(
                            final_response
                        ),
                    )
                }
                stream_start_time = actual_stream_start_time
                await mona_export(False)

            try:
                # Call the actual openai create function without the Mona
                # specific arguments.
                return ResponseGatheringIterator(
                    original_iterator=await inner_super_function(),
                    delta_choice_text_getter=(
                        base_class._get_stream_delta_text_from_choice
                    ),
                    final_choice_getter=base_class._get_final_choice,
                    callback=_stream_done_callback,
                )

            except Exception:
                await inner_handle_exception()
                raise

        @classmethod
        def create(cls, *args, **kwargs):
            """
            A mona-monitored version of the openai base class' "create"
            function.
            """
            return run_in_an_event_loop(
                cls._inner_create(client.export, super().create, args, kwargs)
            )

        @classmethod
        async def acreate(cls, *args, **kwargs):
            """
            An async mona-monitored version of the openai base class'
            "acreate" function.
            """
            return await cls._inner_create(
                async_client.export_async, super().acreate, args, kwargs
            )

        @classmethod
        def get_mona_clients(cls):
            """
            Returns the two Mona clients this class works with to allow
            exporting more data or communicating directly with Mona's
            API.
            """
            return (client, async_client)

    return type(base_class.__name__, (MonitoredOpenAI,), {})


def get_rest_monitor(
    openai_endpoint_name,
    mona_creds,
    context_class,
    specs=EMPTY_DICT,
    mona_clients_getter=get_mona_clients,
):
    """
    Returns a client class for monitoring OpenAI REST calls not done
    using the OpenAI python client (e.g., for Azure users using their
    endpoints directly). This isn't a wrapper for any http requesting
    library and doesn't call the OpenAI API for you - it's just an easy
    logging client to log requests, responses and exceptions.
    """

    client, async_client = mona_clients_getter(mona_creds)

    sampling_ratio = validate_and_get_sampling_ratio(specs)

    wrapping_logic = get_endpoint_wrapping(openai_endpoint_name, specs)

    class RestClient:
        """
        This will be the returned Mona logging class. We follow
        OpenAI's way of doing things by using a static classe with
        relevant class methods.
        """

        @classmethod
        def _inner_log_request(
            cls,
            mona_export_function,
            request_dict,
            additional_data=None,
            context_id=None,
            export_timestamp=None,
        ):
            """
            Actual logic for logging requests, responses and exceptions.
            """
            start_time = time.time()

            inner_response = None

            def _inner_mona_export(is_exception):
                return mona_export_function(
                    _get_mona_single_message(
                        api_name=openai_endpoint_name,
                        request_input=request_dict,
                        start_time=start_time,
                        is_exception=is_exception,
                        is_async=False,
                        # TODO(itai): Support stream in REST as well.
                        stream_start_time=None,
                        response=inner_response,
                        analysis_getter=wrapping_logic.get_full_analysis,
                        context_class=context_class,
                        message_cleaner=wrapping_logic.get_clean_message,
                        additional_data=additional_data,
                        context_id=context_id,
                        export_timestamp=export_timestamp,
                    )
                )

            mona_export = add_conditional_sampling(
                _inner_mona_export, sampling_ratio
            )

            def log_response(response):
                """
                Only when this function is called, will data be logged
                out to Mona. This function should be called with a
                response object from the OpenAI API as close as
                possible to when it is received to allow accurate
                latency logging.
                """
                nonlocal inner_response
                inner_response = response
                return mona_export(False)

            def log_exception():
                return mona_export(True)

            return log_response, log_exception

        @classmethod
        def log_request(
            cls,
            request_dict,
            additional_data=None,
            context_id=None,
            export_timestamp=None,
        ):
            """
            Sets up mona logging for OpenAI request/response objects.

            This function should be called with a request data dict,
            for example, what you would use as "json" when using
            "requests" to post.

            It returns a response logging function to be used with the
            response object, as well as an exception logging function in case
            of exceptions.

            Note that this call does not log anything to Mona until one of the
            returned callbacks is called.
            """
            return cls._inner_log_request(
                client.export,
                request_dict,
                additional_data,
                context_id,
                export_timestamp,
            )

        @classmethod
        def async_log_request(
            cls,
            request_dict,
            additional_data=None,
            context_id=None,
            export_timestamp=None,
        ):
            """
            Async version of "log_request". See function's docstring for more
            details.
            """
            return cls._inner_log_request(
                async_client.export_async,
                request_dict,
                additional_data,
                context_id,
                export_timestamp,
            )

        @classmethod
        def get_mona_clients(cls):
            """
            Returns the two Mona client this class works with to allow
            exporting more data or communicating directly with Mona's
            API.
            """
            return client, async_client

    return RestClient



"""
A module with utility (usually higher order) functions for enriching
and extending other functions' functionalities.
"""
from random import random


def add_conditional_sampling(inner_func, sampling_ratio):
    """
    A higher order function that returns a "sampled" version of the
    given inner function only if needed. This allows for adding
    sampling mechanisms while avoiding conditionals or random number
    creations when either is not necessary.
    """

    async def _sampled_func(*args, **kwargs):
        if random() < sampling_ratio:
            return await inner_func(*args, **kwargs)

    return inner_func if sampling_ratio == 1 else _sampled_func



"""
Logic to create profanity analysis.
"""

from profanity_check import predict_prob, predict

_DECIMAL_PLACES = 2


def get_profanity_prob(texts):
    return tuple(round(x, _DECIMAL_PLACES) for x in predict_prob(texts))


def get_has_profanity(texts):
    return tuple(bool(x) for x in predict(texts))



from mona_openai.analysis.textual import TextualAnalyzer


def test_get_length():
    text = "bla balskdf nblaskd"
    assert TextualAnalyzer(text).get_length() == len(text)


def test_get_word_count():
    text = "bla balskdf nblaskd"
    assert TextualAnalyzer(text).get_word_count() == 3


def test_get_preposition_count():
    text = "bla balskdf of nblaskd from there"
    assert TextualAnalyzer(text).get_preposition_count() == 2


def test_get_preposition_ratio():
    text = "bla balskdf of nblaskd from there"
    assert TextualAnalyzer(text).get_preposition_ratio() == 2 / 6


def test_new_words():
    assert (
        TextualAnalyzer("this is a word").get_words_not_in_others_count(
            (
                TextualAnalyzer("has this word"),
                TextualAnalyzer("no bla bla is"),
            )
        )
        == 1
    )



"""
A utility module for everything realted to encoding tokens.
"""
import tiktoken


def _get_number_of_tokens(text, enc):
    return len(enc.encode(text))


def _get_encoding(model):
    return tiktoken.encoding_for_model(model)


def get_usage(model, prompt_texts, response_texts):
    """
    Returns a usage dict containing the number of tokens in the prompt, in the
    response, and totally.
    """
    enc = _get_encoding(model)

    def get_tokens_sum(texts):
        return sum(_get_number_of_tokens(text, enc) for text in texts)

    usage = {
        "prompt_tokens": get_tokens_sum(prompt_texts),
        "completion_tokens": get_tokens_sum(response_texts),
    }
    usage["total_tokens"] = usage["prompt_tokens"] + usage["completion_tokens"]

    return usage



from mona_openai.analysis.privacy import PrivacyAnalyzer


def test_phone_numbers_count():
    assert (
        PrivacyAnalyzer(
            "Here's a phone number: +972584932014 and another one: "
            "(212)456-7890"
        ).get_phone_numbers_count()
        == 2
    )


def test_emails_count():
    assert (
        PrivacyAnalyzer(
            "Here's an email: itai@monalabs.io and another one: "
            "flsdakjflkjsa@gmail.com"
        ).get_emails_count()
        == 2
    )


def test_other_phone_numbers():
    assert (
        PrivacyAnalyzer(
            "Here's a phone number: +972584932014 and another one: "
            "(212)456-7890"
        ).get_previously_unseen_phone_numbers_count(
            (
                PrivacyAnalyzer(
                    "Here's a different phone number: +972584332014 an "
                    "existing one: (212)456-7890"
                ),
                PrivacyAnalyzer(" a different phone number: +972584332015"),
            )
        )
        == 1
    )


def test_other_emails():
    assert (
        PrivacyAnalyzer(
            "Here's an email: itai@monalabs.io and another one: "
            "flsdakjflkjsa@gmail.com"
        ).get_previously_unseen_emails_count(
            (
                PrivacyAnalyzer(
                    "Here's a different email: flasdkjfg@gmail.com an "
                    "existing one: flsdakjflkjsa@gmail.com"
                ),
                PrivacyAnalyzer(
                    " a different email: f4234lsdakjflkjsa@gmail.com"
                ),
            )
        )
        == 1
    )



"""
A util module for everything related to supporting streams.
"""
import time
from .async_util import run_in_an_event_loop
import inspect


class ResponseGatheringIterator:
    """
    A generator class that takes an original OpenAI stream response generator
    and wraps it with functionality to gather all the stream of responses as
    they come, and create from them a singular reponse object as would have
    been received in non-stream OpenAI usage.

    Once the original generator is done it creates the full response and calls
    a callback with it.

    It acts both as sync and async generator to ease the use of sync/async
    joint code.
    """

    def __init__(
        self,
        delta_choice_text_getter,
        final_choice_getter,
        original_iterator,
        callback,
    ):
        self._original_iterator = original_iterator
        self._delta_choice_text_getter = delta_choice_text_getter
        self._final_choice_getter = final_choice_getter
        self._callback = callback
        self._initial_event_recieved_time = None
        self._common_response_information = None
        self._choices = {}

    def __iter__(self):
        return self

    def __aiter__(self):
        return self

    def __next__(self):
        try:
            return self._add_response(self._original_iterator.__next__())
        except StopIteration:
            self._call_callback()
            raise

    async def __anext__(self):
        try:
            return self._add_response(
                await self._original_iterator.__anext__()
            )
        except StopAsyncIteration:
            await self._a_call_callback()
            raise

    def _add_response(self, event):
        """
        The main and only exposed function of the ResponseGatherer class. Use
        this function to collect stream events.
        """
        if self._initial_event_recieved_time is None:
            self._initial_event_recieved_time = time.time()
            self._common_response_information = {
                x: event[x] for x in event if x != "choices"
            }

        # Gather response events by choice index.
        self._handle_choice(self._get_only_choice(event))
        return event

    def _call_callback(self):
        # We allow an async function as the callback event if this class is
        # used as a sync generator. This code handles this scenario.
        callback_args = (
            self._create_singular_response(),
            self._initial_event_recieved_time,
        )
        if inspect.iscoroutinefunction(self._callback):
            run_in_an_event_loop(self._callback(*callback_args))
            return

        self._callback(*callback_args)

    async def _a_call_callback(self):
        await self._callback(
            self._create_singular_response(),
            self._initial_event_recieved_time,
        )

    def _handle_choice(self, choice):
        index = choice["index"]
        self._choices[index] = self._choices.get(index, []) + [choice]

    def _get_only_choice(self, event):
        # Stream response events have only a single choice that specifies
        # its own index.
        return event["choices"][0]

    def _create_singular_response(self):
        choices = [
            self._get_full_choice(choice) for choice in self._choices.values()
        ]
        return self._common_response_information | {"choices": choices}

    def _get_full_choice(self, choice):
        full_text = "".join(
            self._delta_choice_text_getter(choice_event)
            for choice_event in choice
        )

        return {
            **self._final_choice_getter(full_text),
            "index": choice[0]["index"],
            "finish_reason": choice[-1]["finish_reason"],
        }



"""
A module for general logic for wrapping OpenAI endpoints.
"""
import abc
from ..util.validation_util import validate_openai_class


class OpenAIEndpointWrappingLogic(metaclass=abc.ABCMeta):
    """
    An abstract class used for wrapping OpenAI endpoints. Each child of this
    class must implement several key logics for wrapping specific OpenAI
    endpoints' logic.

    Note that subclasses are not the actual wrappers of the OpenAI endpoint
    classes, but provide functions to create the actual wrapper classes.
    """

    def __init__(self, specs):
        self._specs = specs
        self._analysis_functions = {
            "privacy": self._get_full_privacy_analysis,
            "textual": self._get_full_textual_analysis,
            "profanity": self._get_full_profainty_analysis,
        }

    def wrap_class(self, openai_class):
        """
        Returns a monitored class wrapping the given openai class, enriching it
        with specific capabilities to be used by an inhereting monitored class.
        """
        validate_openai_class(openai_class, self._get_endpoint_name())

        class WrapperClass(openai_class):
            # TODO(itai): Have a smarter way to "import" all the methods to
            #   this class instead of just copying them.
            @classmethod
            def _get_full_analysis(cls, input, response):
                return self.get_full_analysis(input, response)

            @classmethod
            def _get_clean_message(cls, message):
                return self.get_clean_message(message)

            @classmethod
            def _get_stream_delta_text_from_choice(cls, choice):
                return self.get_stream_delta_text_from_choice(choice)

            @classmethod
            def _get_final_choice(cls, text):
                return self.get_final_choice(text)

            @classmethod
            def _get_all_prompt_texts(cls, request):
                return self.get_all_prompt_texts(request)

            @classmethod
            def _get_all_response_texts(cls, response):
                return self.get_all_response_texts(response)

        return type(
            f"Monitored{self._get_endpoint_name()}", (WrapperClass,), {}
        )

    @abc.abstractmethod
    def _get_endpoint_name(self):
        """
        Returns the name of the OpenAI endpoint that is being wrapped.
        """
        pass

    @abc.abstractmethod
    def get_clean_message(self, message):
        """
        Given a mona message, returns a "clean" message in the sense that it
        will not hold any information that shouldn't be exported to Mona
        (e.g., actual prompts).
        """
        pass

    def get_full_analysis(self, input, response):
        """
        Returns a dict mapping each analysis type to all related analysis
        fields for the given prompt and answers according to the given
        specs (if no "analysis" spec is given - return result for all
        analysis types).

        TODO(itai): Consider propogating the specs to allow the user to
            choose specific anlyses to be made from within each analysis
            category.
        """
        return {
            x: self._analysis_functions[x](input, response)
            for x in self._analysis_functions
            if self._specs.get("analysis", {}).get(x, True)
        }

    @abc.abstractmethod
    def _get_full_privacy_analysis(self, input, response):
        """
        Returns a dictionary with all calculated privacy analysis params.
        """
        pass

    @abc.abstractmethod
    def _get_full_textual_analysis(self, input, response):
        """
        Returns a dictionary with all calculated textual analysis params.
        """
        pass

    @abc.abstractmethod
    def _get_full_profainty_analysis(self, input, response):
        """
        Returns a dictionary with all calculated profanity analysis params.
        """
        pass

    @abc.abstractclassmethod
    def get_stream_delta_text_from_choice(self, choice):
        """
        Given a stream response "choice", returns the text from that choice.
        """
        pass

    @abc.abstractclassmethod
    def get_final_choice(self, text):
        """
        Returns a dictionary for a "choice" object as it would have been
        received from OpenAI's API that holds the given text as the content.
        """
        pass

    @abc.abstractclassmethod
    def get_all_prompt_texts(self, request):
        """
        Given a request object, returns all the prompt texts from that
        request.
        """
        pass

    @abc.abstractclassmethod
    def get_all_response_texts(self, response):
        """
        Given a response object, returns all the possible response texts.
        """
        pass



"""
Utility logic for object-oriented stuff.
"""
import inspect


def create_combined_object(instances):
    """
    Create a new object that has the same methods as the given instances.

    This function takes an iterable of instances of a given class and returns
    a new object that has the same methods as the given instances. When calling
    these methods, it returns a tuple containing all the results of running
    that method for all instances.

    Args:
        instances: An iterable of instances of a given class.

    Returns:
        A new object that has the same methods as the given instances.
    """

    class CombinedObject:
        def __init__(self, instances):
            self._instances = instances

        def __getattr__(self, name):
            def method(*args, **kwargs):
                results = []
                for instance in self._instances:
                    func = getattr(instance, name)
                    if inspect.ismethod(func) or inspect.isfunction(func):
                        results.append(func(*args, **kwargs))
                return tuple(results)

            return method

    return CombinedObject(instances)



"""
Util logic for OpenAI related logic and language (e.g., dealing with API
parameter names).
"""


def get_model_param(request):
    """
    Returns the "model" param in the request, the "engine" param if no
    "model" param is used, and None if neither exists (which isn't expected
    to happen)
    """
    return request.get("model", request.get("engine"))



import asyncio
from types import MappingProxyType
from inspect import iscoroutinefunction


def run_in_an_event_loop(coroutine):
    """
    A light wrapper around asyncio.run to avoid crushing when trying to run a
    coroutine in an environment where an event loop is already in place and
    asyncio.run doesn't work.
    """
    try:
        return asyncio.run(coroutine)
    except RuntimeError:
        return asyncio.get_event_loop().run_until_complete(coroutine)


EMPTY_DICT = MappingProxyType({})


async def call_non_blocking_sync_or_async(
    function, func_args=(), func_kwargs=EMPTY_DICT
):
    """
    A higher order function that allows calling both sync and async
    functions as if they were async, avoid blocking when relevant, and
    maintain one code base for both cases.
    """
    if iscoroutinefunction(function):
        return await function(*func_args, **func_kwargs)
    return function(*func_args, **func_kwargs)
